#%%
""" 
如果要切換訓練或測試模型，可以修改354行的mission = 'train'
其中'train'改成'test'就是測試模型用的了
"""

import joblib
import pickle
import os
import numpy as np
import pandas as pd 

# 基於特徵
from adapt.feature_based import PRED
from adapt.feature_based import FA
from adapt.feature_based import SA
from adapt.feature_based import DANN
from adapt.feature_based import MDD
from adapt.feature_based import WDGRL

# 基於實例
from adapt.instance_based import LDM
from adapt.instance_based import KLIEP
from adapt.instance_based import NearestNeighborsWeighting
from adapt.instance_based import IWC
from adapt.instance_based import BalancedWeighting
from adapt.instance_based import TrAdaBoostR2
from adapt.instance_based import TwoStageTrAdaBoostR2
from adapt.instance_based import WANN

from adapt.utils import make_classification_da
from sklearn.linear_model import LogisticRegression
from sklearn.linear_model import RidgeClassifier
from sklearn.linear_model import Ridge
from sklearn.svm import SVR
from sklearn.linear_model import LinearRegression
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.gaussian_process.kernels import Matern, WhiteKernel

import tensorflow as tf
from tensorflow.keras.callbacks import Callback

import seaborn as sns
import matplotlib.pyplot as plt

import time

from functools import wraps

import glob
import traceback

# 初始化設定
X_list = [
    # 原圖顏色特徵17個
    'R','G','B','H1','S1','V1','L2','S2','L','a','b','Y','Cr','Cb','NDI','GI','RGRI',
    # 校正後顏色特徵17個
    'RC','GC','BC','H1C','S1C','V1C','L2C','S2C','LC','aC','bC','YC','CrC','CbC','NDIC','GIC','RGRIC',
    # 原本的模型(只有校正板的平均數值沒有RGB數值
    # 'B_CB',	'G_CB',	'W_CB',
    # 後來的模型(有校正板的RGB數值
    'B_RGB_B', 'B_RGB_G', 'B_RGB_R',
    'G_RGB_B', 'G_RGB_G', 'G_RGB_R',
    'W_RGB_B', 'W_RGB_G', 'W_RGB_R']

Y_list = ['HMC']

def timeit(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        start_time = time.time()
        # 假设第一个参数总是model_name
        model_name = args[0] if args else 'Unknown Model'
        result = func(*args, **kwargs)
        end_time = time.time()
        print(f"{model_name} 訓練完成，花費時間：{(end_time - start_time)/60:.2f} 分鐘。")
        return result
    return wrapper

def save_model(model):
    try:
        os.makedirs(f'./02_Model_saved/{model_name},n={target_number},epochs={epochs}', exist_ok=True)
        model.save(f'./02_Model_saved/{model_name},n={target_number},epochs={epochs}')
    except:
        if os.path.isdir(f'./02_Model_saved/{model_name},n={target_number},epochs={epochs}')==True:os.rmdir(f'./02_Model_saved/{model_name},n={target_number},epochs={epochs}')
        # To save a model
        joblib.dump(model, f'./02_Model_saved/{model_name},n={target_number},epochs={epochs}.pkl')

st_time = time.time()

random_state = 52024101
epochs=1000
target_number = 10

def write_to_file(result_name, image_names, truth, predict, periods):
    # 確保數組長度一致
    if len(image_names) == len(predict):
        # 轉換為 DataFrame
        df_image_names = pd.DataFrame(image_names, columns=['image_name'])
        df_truth = pd.DataFrame(truth, columns=['truth'])
        df_predict = pd.DataFrame(predict, columns=['predict'])
        df_error = pd.DataFrame(abs(df_truth['truth'] - df_predict['predict']), columns=['error'])

        # 合併 DataFrame
        df = pd.concat([df_image_names, df_truth, df_predict, df_error], axis=1)

        # 新增計算不同期間的MAE
        df_periods = pd.DataFrame(periods, columns=['Period'])
        df = pd.concat([df_image_names, df_truth, df_predict, df_error, df_periods], axis=1)  # 新增df_periods合併

        # 分別計算1121和1122期的MAE
        mae_112 = round(df[(df['Period']==1121) | (df['Period']==1122)]['error'].mean(), 2)
        mae_1121 = round(df[df['Period'] == 1121]['error'].mean(), 2)
        mae_1122 = round(df[df['Period'] == 1122]['error'].mean(), 2)
        # 更新檔案名稱以包含MAE信息
        csv_path = f'./03_predict_result/{result_name}_112={mae_112}_1121={mae_1121}_1122={mae_1122}.csv'
        df.to_csv(csv_path, index=False)
        # 繪製散佈圖
        sns.scatterplot(x='truth', y='predict', data=df)
        plt.xlabel('Truth')
        plt.ylabel('Predict')
        plt.title(f'{result_name}')
        plt.xlim(0,80)
        plt.ylim(0,80)
        # 儲存散佈圖
        plt.savefig(csv_path.replace('.csv', '.png'))
        plt.close()


        #計算含水量35%以下的MAE
        df_below35 = df[(df['truth'])<35]
        # 分別計算1121和1122期的MAE
        mae_112 = round(df_below35[(df_below35['Period']==1121) | (df_below35['Period']==1122)]['error'].mean(), 2)
        mae_1121 = round(df_below35[df_below35['Period'] == 1121]['error'].mean(), 2)
        mae_1122 = round(df_below35[df_below35['Period'] == 1122]['error'].mean(), 2)
        # 更新檔案名稱以包含MAE信息
        below35_csv_path = f'./03_predict_result/below35_{result_name}_112={mae_112}_1121={mae_1121}_1122={mae_1122}.csv'
        df_below35.to_csv(below35_csv_path, index=False)

    else:
        print("數組長度不一致，請檢查數據。")

def data_init():
    df = pd.read_csv(r'D:\RiceProject\3_rice_moisture\01_raw_data\1082-1122期作_9446筆_穀粒含水量調查四年彙整資料_剔除不適用與含水量60以上_新的.csv')

    train_df = df[(df['Period']!=1121) & (df['Period']!=1122)]
    SDA_df = df[(df['Period']==1121) | (df['Period']==1122)]
    SDA_df = SDA_df.sample(n=target_number, random_state=random_state) #10 100 908
    # print(train_df.shape)
    # print(test_df.shape)

    dt = pd.read_csv(r'D:\RiceProject\3_rice_moisture\01_raw_data\1121期作_2430筆_沒有含水量真值的資料_AIpal小精靈_新的.csv')
    test_df = dt[(dt['ID']=='NGT')]
    test_df = test_df.sample(n=target_number, random_state=random_state) #10 100 2430

    periods_s = train_df['Period'].values  # 新增 - 源期數據
    periods_t = test_df['Period'].values  # 新增 - 目標期數據

    Xs = train_df[X_list].values
    ys = train_df[Y_list].values

    Xt = test_df[X_list].values
    # ! 因為目標領域沒有真值，所以刪除
    # yt = test_df[Y_list].values

    X_SDA = SDA_df[X_list].values
    y_SDA = SDA_df[Y_list].values

    print('train',Xs.shape)
    print('UDA',Xt.shape)
    print('SDA',X_SDA.shape)

    image_names_s = train_df['image_name'].values
    image_names_t = test_df['image_name'].values

    return Xs, ys, Xt, image_names_s, image_names_t, periods_s, periods_t, X_SDA, y_SDA

if __name__ == '__main__':
    Xs, ys, Xt, image_names_s, image_names_t, periods_s, periods_t, X_SDA, y_SDA = data_init()
    model_name_list = [
        ##基於特徵 (Feature-based)
        'SDA_Feature-based_PRED',
        'SDA_Feature-based_FA',
        'UDA_Feature-based_SA',
        'UDA_Feature-based_DANN',   
        'UDA_Feature-based_WDGRL',    
        'UDA_Feature-based_MDD', 

        # ##基於實例 (Instance-based)
        'UDA_Instance-based_LDM',   
        'UDA_Instance-based_KLIEP',
        'UDA_Instance-based_NNW',
        'UDA_Instance-based_IWC',
        'SDA_Instance-based_BW',
        'SDA_Instance-based_TABR2',
        'SDA_Instance-based_TSTABR2',
        'SDA_Instance-based_WANN', 
    ]

    @timeit
    def train_and_evaluate(model_name, Xs, Xt, image_names_t, periods_t, X_SDA, y_SDA):
        df = pd.read_csv(r'D:\RiceProject\3_rice_moisture\01_raw_data\1082-1122期作_9446筆_穀粒含水量調查四年彙整資料_剔除不適用與含水量60以上_新的.csv')
        test_df = df[(df['Period']==1121) | (df['Period']==1122)]
        X_test = test_df[X_list].values
        y_test = test_df[Y_list].values
        image_names_test = test_df['image_name'].values
        periods_test = test_df['Period'].values
        # 檢查該模型是否已經跑過，如果跑過就return結束
        print(f'開始訓練 {model_name}')
        search_checker = len(glob.glob(f'./02_Model_saved/{model_name},n={target_number},epochs={epochs}')) >0
        if search_checker:
            print(f'{model_name}  已經訓練過了')
            return
        #========PRED========
        if model_name == "SDA_Feature-based_PRED":
            model = PRED(SVR(kernel='rbf',C=epochs,gamma='scale',epsilon=0.1), Xt=X_SDA[[1, -1, -2]], yt=y_SDA[[1, -1, -2]],pretrain=True, verbose=0, random_state=random_state)
            model.fit(Xs, ys)
            save_model(model)
            write_to_file(f'{model_name},n={target_number},epochs={epochs}', image_names_test, y_test, model.predict(X_test), periods_test)
        #========PRED========

        #========FA========
        elif model_name == "SDA_Feature-based_FA":
            #數值越小越好
            model = FA(SVR(kernel='rbf',C=epochs,gamma='scale',epsilon=0.1), Xt=X_SDA, yt=y_SDA, random_state=random_state)
            model.fit(Xs, ys)
            save_model(model)
            write_to_file(f'{model_name},n={target_number},epochs={epochs}', image_names_test, y_test, model.predict(X_test), periods_test)
        #========FA========

        #=======SA=======
        elif model_name == "UDA_Feature-based_SA":
            model = SA(SVR(kernel='rbf',C=epochs,gamma='scale',epsilon=0.1), Xt=Xt, random_state=random_state)
            model.fit(Xs, ys)
            save_model(model)
            write_to_file(f'{model_name},n={target_number},epochs={epochs}', image_names_test, y_test, model.predict(X_test), periods_test)
        #=======SA=======

        #=======DANN=======
        elif model_name=='UDA_Feature-based_DANN':
            model = DANN(lambda_=0.1, Xt=Xt, metrics=["mae"], random_state=random_state)
            model.fit(Xs, ys, epochs=epochs, verbose=0)
            save_model(model)
            write_to_file(f'{model_name},n={target_number},epochs={epochs}', image_names_test, y_test, model.predict(X_test), periods_test)
        #=======DANN=======

        #=======WDGRL=======
        elif model_name=='UDA_Feature-based_WDGRL':
            model = WDGRL(lambda_=1., gamma=1., Xt=Xt, metrics=["mae"], random_state=random_state)
            model.fit(Xs, ys, epochs=epochs, verbose=0)
            save_model(model)
            write_to_file(f'{model_name},n={target_number},epochs={epochs}', image_names_test, y_test, model.predict(X_test), periods_test)
        #=======WDGRL=======

        #=======MDD=======
        elif model_name=='UDA_Feature-based_MDD':
            model = MDD(lambda_=0.1, gamma=4., Xt=Xt, metrics=["mae"], random_state=random_state)
            model.fit(Xs, ys, epochs=epochs, verbose=0)
            save_model(model)
            write_to_file(f'{model_name},n={target_number},epochs={epochs}', image_names_test, y_test, model.predict(X_test), periods_test)
        #=======MDD=======

        # 基於實例的方法
        #=======LDM=======
        elif model_name=='UDA_Instance-based_LDM':
            model = LDM(SVR(kernel='rbf',C=epochs,gamma='scale',epsilon=0.1), Xt=Xt, random_state=random_state)
            model.fit(Xs, ys)
            save_model(model)
            write_to_file(f'{model_name},n={target_number},epochs={epochs}', image_names_test, y_test, model.predict(X_test), periods_test)
        #=======LDM=======

        #=======KLIEP=======
        elif model_name=='UDA_Instance-based_KLIEP':
            model = KLIEP(SVR(kernel='rbf',C=epochs,gamma='scale',epsilon=0.1), Xt=Xt, kernel="rbf", random_state=random_state)
            model.fit(Xs, ys)
            save_model(model)
            write_to_file(f'{model_name},n={target_number},epochs={epochs}', image_names_test, y_test, model.predict(X_test), periods_test)
        #=======KLIEP=======

        #=======NearestNeighborsWeighting=======
        elif model_name=='UDA_Instance-based_NNW':
            model = NearestNeighborsWeighting(SVR(kernel='rbf',C=epochs,gamma='scale',epsilon=0.1), n_neighbors=5, Xt=Xt, random_state=random_state)
            model.fit(Xs, ys)
            save_model(model)
            write_to_file(f'{model_name},n={target_number},epochs={epochs}', image_names_test, y_test, model.predict(X_test), periods_test)
        #=======NearestNeighborsWeighting=======

        #=======IWC=======
        elif model_name=='UDA_Instance-based_IWC':
            model = IWC(SVR(kernel='rbf',C=epochs,gamma='scale',epsilon=0.1), classifier=RidgeClassifier(0.), Xt=Xt, random_state=random_state)
            model.fit(Xs, ys)
            save_model(model)
            write_to_file(f'{model_name},n={target_number},epochs={epochs}', image_names_test, y_test, model.predict(X_test), periods_test)
        #=======IWC=======

        #=======BalancedWeighting=======
        elif model_name=='SDA_Instance-based_BW':
            model = BalancedWeighting(SVR(kernel='rbf',C=epochs,gamma='scale',epsilon=0.1), gamma=0.5, Xt=X_SDA, yt=y_SDA, verbose=0, random_state=random_state)
            model.fit(Xs, ys)
            save_model(model)
            write_to_file(f'{model_name},n={target_number},epochs={epochs}', image_names_test, y_test, model.predict(X_test), periods_test)
        #=======BalancedWeighting=======

        #=======TrAdaBoostR2=======
        elif model_name=='SDA_Instance-based_TABR2':
            model = TrAdaBoostR2(SVR(kernel='rbf',C=epochs,gamma='scale',epsilon=0.1), n_estimators=10, Xt=X_SDA, yt=y_SDA, random_state=random_state)
            model.fit(Xs, ys)
            save_model(model)
            write_to_file(f'{model_name},n={target_number},epochs={epochs}', image_names_test, y_test, model.predict(X_test), periods_test)
        #=======TrAdaBoostR2=======

        #=======TwoStageTrAdaBoostR2=======
        elif model_name=='SDA_Instance-based_TSTABR2':
            model = TwoStageTrAdaBoostR2(SVR(kernel='rbf',C=epochs,gamma='scale',epsilon=0.1), n_estimators=10, Xt=X_SDA, yt=y_SDA, random_state=random_state)
            model.fit(Xs, ys)
            save_model(model)
            write_to_file(f'{model_name},n={target_number},epochs={epochs}', image_names_test, y_test, model.predict(X_test), periods_test)
        #=======TwoStageTrAdaBoostR2=======

        #=======WANN=======
        elif model_name=='SDA_Instance-based_WANN':
            model = WANN(Xt=X_SDA, yt=y_SDA, random_state=random_state)
            model.fit(Xs, ys, epochs=epochs, verbose=0)
            save_model(model)
            write_to_file(f'{model_name},n={target_number},epochs={epochs}', image_names_test, y_test, model.predict(X_test), periods_test)
        #=======WANN=======

    def model_test(model_name, Xt, image_names_t):
        try:
            model_path = glob.glob(f'./02_Model_saved/*{model_name}*')[0]
        except:return

        try:
            if not '.pkl' in model_path:
                model = tf.keras.models.load_model(model_path)
            else:
                # 使用pickle加载模型
                with open(model_path, 'rb') as file:
                    model = pickle.load(file)
        except:return

        predictions = model.predict(Xt)
        print(f'model_name={model_name}, predictions={np.mean(predictions):.2f}, predictions.shape={predictions.shape}')

        df_image_names = pd.DataFrame(image_names_t, columns=['image_name'])
        df_GMC = pd.DataFrame(predictions, columns=[f'Predict_GMC_{model_name}'])

        # 合併 DataFrame
        df = pd.concat([df_image_names, df_GMC], axis=1)
        # 更新檔案名稱以包含MAE信息
        csv_path = f'./Result/{model_name}.csv'
        df.to_csv(csv_path, index=False)

    mission = 'train'

    for model_name in model_name_list:
        if mission == 'train':
            try:
                train_and_evaluate(model_name, Xs, Xt, image_names_t, periods_t, X_SDA, y_SDA)
            except:
                print(traceback.format_exc())
        else:
            try:
                model_test(model_name, Xt, image_names_t)
            except:
                print(traceback.format_exc())

print(f'訓練花費時間 {round((time.time() - st_time) / 60, 2)} 分鐘')