#%%
import glob 
from tqdm import trange
from pathlib import Path
import pandas as pd
import csv
import joblib

import seaborn as sns
import matplotlib.pyplot as plt
#==================================驗證資料=============================
X_list = [
    # 原圖顏色特徵17個
    'R','G','B','H1','S1','V1','L2','S2','L','a','b','Y','Cr','Cb','NDI','GI','RGRI',
    # 校正後顏色特徵17個
    'RC','GC','BC','H1C','S1C','V1C','L2C','S2C','LC','aC','bC','YC','CrC','CbC','NDIC','GIC','RGRIC',
    # 原本的模型(只有校正板的平均數值沒有RGB數值
    # 'B_CB',	'G_CB',	'W_CB',
    # 後來的模型(有校正板的RGB數值
    'B_RGB_B', 'B_RGB_G', 'B_RGB_R',
    'G_RGB_B', 'G_RGB_G', 'G_RGB_R',
    'W_RGB_B', 'W_RGB_G', 'W_RGB_R']

path = './02_Model_saved'
TESTperiod = 1121
target_model = glob.glob(f'{path}/02_Model_saved/*.pkl')
print(target_model)

for i in range(len(target_model)):
    target_model_path = glob.glob(f'{path}/*.pkl')[i]
    target_model_name = Path(target_model_path).stem
    print(target_model_name)

    def HMC_estimate(X):
        #读取Model
        DA = joblib.load(target_model_path)
        X = [X]
        DA_prediction = DA.predict(X)
        HMC_ = round(DA_prediction[0],1)
        return HMC_

    test_file_path = r'D:\RiceProject\3_rice_moisture\01_raw_data\1082-1122期作_9446筆_穀粒含水量調查四年彙整資料_剔除不適用與含水量60以上_新的.csv'
    df = pd.read_csv(test_file_path)
    # # 找到符合條件的行索引
    rows_to_remove = df[df['Period'] != TESTperiod].index
    # # 刪除這些行
    test_df = df.drop(rows_to_remove)


    #print(df)
    with open(f'{path}/preictions_by_{target_model_name}.csv', 'w', encoding='utf-8-sig', newline='') as df:
        writer = csv.writer(df)
        writer.writerow(['image_name','Period', 'truth', 'predict', 'error'])
    
    with open(f'{path}/preictions_by_{target_model_name}.csv', 'a', encoding='utf-8-sig', newline='') as df:
        writer = csv.writer(df)
        all_error = []
        for i in trange(len(test_df), colour='blue'):
            this_data = test_df.iloc[i]
            image_name = this_data['image_name']
            Period = this_data['Period']
            HMC = this_data['HMC']
            X = this_data[X_list].values
            predict = HMC_estimate(X)
            error = round(abs(HMC - predict), 2)
            all_error.append(error)
            # print(image_name, HMC, predict, error)
            writer.writerow([image_name,Period, HMC, predict, error])
        MAE = sum(all_error) / len(all_error)
        print(MAE)

    preictions_path = glob.glob(f'{path}/preictions_by_{target_model_name}.csv')
    print(preictions_path)
    df =  pd.concat(map(pd.read_csv, preictions_path))

    # 計算MAE
    mae_TESTperiod = round(df[df['Period'] == TESTperiod]['error'].mean(), 2)
    # 更新檔案名稱以包含MAE信息
    csv_path = f'{path}/{target_model_name}_{TESTperiod}={mae_TESTperiod}.csv'
    df.to_csv(csv_path, index=False)
    # 繪製散佈圖
    sns.scatterplot(x='truth', y='predict', data=df)
    plt.xlabel('Truth')
    plt.ylabel('Predict')
    plt.title(f'{target_model_name}')
    plt.xlim(0,80)
    plt.ylim(0,80)
    # 儲存散佈圖
    plt.savefig(csv_path.replace('.csv', '.png'))
    plt.close()

    #計算含水量35%以下的MAE
    df_below35 = df[(df['truth'])<35]
    # 分別計算1121和1122期的MAE
    mae_TESTperiod = round(df_below35[df_below35['Period'] == TESTperiod]['error'].mean(), 2)
    # 更新檔案名稱以包含MAE信息
    below35_csv_path = f'{path}/below35_{target_model_name}_{TESTperiod}={mae_TESTperiod}.csv'
    df_below35.to_csv(below35_csv_path, index=False)